 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Debugger
  * 
  * (c) 1995 Bernd Schmidt
  * 
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <signal.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
#include "options.h"
#include "debug.h"
#include "memory.h"
#include "custom.h"
#include "newcpu.h"

void Quit_vMac(void);

void fatal_error (char *fmt, ...)
{
  	va_list args;

  	va_start (args, fmt);
  	printf("vMac Fatal Error: ");
  	vprintf(fmt, args);
  	printf("\n");
  	va_end(args);

  	Quit_vMac();
}


static int caught_fatal_sig = 0;

static RETSIGTYPE on_signal (int sig_num)
{
  	char *sig;

  	if (caught_fatal_sig)
    		kill (getpid (), sig_num);
  	caught_fatal_sig = 1;

  	switch (sig_num)
    	{
    		case SIGHUP:
      			sig = "sighup";
      			break;    
      		case SIGINT:
      			sig = "sigint";
      			break;
    		case SIGQUIT:
      			sig = "sigquit";
      			break;
    		case SIGABRT:
      			sig = "sigabrt";
      			break;
    		case SIGBUS:
      			sig = "sigbus";
      			break;
    		case SIGSEGV:
      			sig = "sigsegv";
      			break;
    		case SIGPIPE:
      			sig = "sigpipe";
      			break;
    		case SIGTERM:
      			sig = "sigterm";
      			break;
    		default:
      			sig = "unknown signal";
      			break;
    	}
  	fatal_error ("%s caught.", sig);
}

void vMac_SignalHandler(void) {

	/* Handle some signals */
  	signal (SIGHUP, on_signal);
  	signal (SIGINT, on_signal);
  	signal (SIGQUIT, on_signal);
  	signal (SIGABRT, on_signal);
  	signal (SIGBUS, on_signal);
  	signal (SIGSEGV, on_signal);
  	signal (SIGPIPE, on_signal);
  	signal (SIGTERM, on_signal);

	printf("vMac_SignalHandler() engaged. We're now protected.\n");

}

static void ignore_ws(char **c)
{
    while (**c && isspace(**c)) (*c)++;
}

static ULONG readhex(char **c)
{
    ULONG val = 0;
    char nc;

    ignore_ws(c);

    while (isxdigit(nc = **c)){
	(*c)++;
	val *= 16;
	nc = toupper(nc);
	if (isdigit(nc)) {
	    val += nc - '0';
	} else {
	    val += nc - 'A' + 10;
	}
    }
    return val;
}

static char next_char(char **c)
{
    ignore_ws(c);
    return *(*c)++;
}

static int more_params(char **c)
{
    ignore_ws(c);
    return (**c) != 0;
}

static void dumpmem(CPTR addr, CPTR *nxmem, int lines)
{
    broken_in = 0;
    for (;lines-- && !broken_in;){
	int i;
	printf("%08lx ", addr);
	for(i=0; i< 16; i++) {
	    printf("%04x ", get_word(addr)); addr += 2;
	}
	printf("\n");
    }
    *nxmem = addr;
}

void debug(void)
{
    char input[80];
    CPTR nextpc,nxdis,nxmem;

#if DJGPP
    signal(SIGINT, sigbrkhandler);
    if (!use_debugger) {
	signal(SIGINT, exit);
	MC68000_run();
    }
#endif
    printf("debugging...\n");
    MC68000_dumpstate(&nextpc);
    nxdis = nextpc; nxmem = 0;
    for(;;){
	char cmd,*inptr;

	printf(">");
	fgets(input, 80, stdin);
	inptr = input;
	cmd = next_char(&inptr);
	switch(cmd){
	 case 'q': 
	    return;
	 case 'r':
	    MC68000_dumpstate(&nextpc);
	    break;
	 case 'd': 
	    {
		ULONG daddr;
		int count;
		
		if (more_params(&inptr))
		    daddr = readhex(&inptr);
		else 
		    daddr = nxdis;
		if (more_params(&inptr))
		    count = readhex(&inptr); 
		else
		    count = 10;
		MC68000_disasm(daddr, &nxdis, count);
	    }
	    break;
	 case 'R':
	 {
	   MC68000_reset();
	   break;
	 }
	 case 'I':
	 {
	   extern int InterruptNo;
	   
	   specialflags |= SPCFLAG_DOINT;
       InterruptNo = 4;
       break;
     }
	 case 't': 
	    MC68000_step(); 
	    MC68000_dumpstate(&nextpc); 
	    break;
	 case 'z': 
	    MC68000_skip(nextpc);
	    MC68000_dumpstate(&nextpc);
	    nxdis = nextpc; 
	    break;
	 case 'f': 
	    MC68000_skip(readhex(&inptr));
	    MC68000_dumpstate(&nextpc); 
	    break;
	 case 'g':
	    if (more_params (&inptr))
		m68k_setpc (readhex (&inptr));
#ifdef DJGPP
	    { extern void EnterGraphics(void); EnterGraphics(); }
#endif
	    MC68000_run();
	    MC68000_dumpstate(&nextpc);
	    break;
	 case 'm':
	    {
		ULONG maddr; int lines;
		if (more_params(&inptr))
		    maddr = readhex(&inptr); 
		else 
		    maddr = nxmem;
		if (more_params(&inptr))
		    lines = readhex(&inptr); 
		else 
		    lines = 16;
		dumpmem(maddr, &nxmem, lines);
	    }
	    break;
          case 'h':
          case '?':
	    {
             printf ("          HELP for vMac Debugger\n");
             printf ("         -----------------------\n\n");
             printf ("  g: <address>          ");
                printf("Start execution at the current address or <address>.\n");
             printf ("  r:                    ");
                printf("Dump state of the CPU\n");
             printf ("  m <address> <lines>:  ");
                printf ("Memory dump starting at <address>\n");
             printf ("  d <address> <lines>:  ");
                printf ("Disassembly starting at <address>\n");
             printf ("  t:                    ");
                printf ("Step one instruction\n");
             printf ("  z:                    ");
                printf ("Step through one instruction - useful for JSR, DBRA etc\n");
             printf ("  f <address>:          ");
                printf ("Step forward until PC == <address>\n");
             printf ("  h,?:                  ");
                printf ("Show this help page\n");
             printf ("  R:                    ");
                printf ("Restart the Macintosh Plus.\n");
             printf ("  I:                    ");
                printf ("Interrupt the Macintosh Plus.\n");
             printf ("  q:                    ");
                printf ("Quit the emulator. You don't want to use this command.\n\n");
            }
            break;
             
	}
    }
}
